/* $Id: cpl_geom_img-test.c,v 1.28 2009/03/11 16:09:04 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2008 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2009/03/11 16:09:04 $
 * $Revision: 1.28 $
 * $Name: cpl-5_0_0 $
 */

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <math.h>

#include <cpl_image_gen.h>
#include <cpl_imagelist_io.h>
#include <cpl_image_io.h>
#include <cpl_image_basic.h>
#include <cpl_bivector.h>
#include <cpl_vector.h>
#include <cpl_memory.h>
#include <cpl_msg.h>
#include <cpl_test.h>
#include <cpl_plot.h>

/* cpl_drand() */
#include <cpl_tools.h>

#include "cpl_apertures.h"
#include "cpl_geom_img.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define     IMAGESZ             256
#define     NFRAMES             10
#define     NSIGMAS              4
#define     MAX_SHIFT_ERROR1    15
#define     MAX_SHIFT_ERROR2    0.1

/*-----------------------------------------------------------------------------
                                  Pricate functions
 -----------------------------------------------------------------------------*/

void cpl_geom_img_offset_saa_bench(cpl_geom_combine, int, int, int, int, int);

/**@{*/

/*-----------------------------------------------------------------------------
                                  Main
 -----------------------------------------------------------------------------*/
int main(void)
{
    const cpl_geom_combine geoms[] = {CPL_GEOM_INTERSECT, CPL_GEOM_UNION,
                                      CPL_GEOM_FIRST};
    const double        off_x_init[] = { 0.0, -6.0, -18.0, 54.0,  33.0,
                                        46.0, -3.0,  36.0, 42.0, -13.0};
    const double        off_y_init[] = { 0.0, 13.0,   3.0, 8.0,  32.0,
                                        22.0, 18.0, -56.0, 3.0,  10.0};
    cpl_imagelist   *   iset;
    cpl_image       *   img;
    cpl_bivector    *   offs_est;
    cpl_vector      *   off_vec_x;
    cpl_vector      *   off_vec_y;
    cpl_bivector    *   offs_ref;
    double          *   x_offs;
    double          *   y_offs;
    cpl_apertures   *   aperts;
    int                 naperts;
    cpl_bivector    *   aperts_pos;
    cpl_vector      *   aperts_pos_x;
    cpl_vector      *   aperts_pos_y;
    cpl_vector      *   correl;
    const double        psigmas[] = {5, 2, 1, 0.5};
    cpl_vector      *   sigmas;
    cpl_image       **  combined;
    int                 i;
    int                 pos;

    cpl_test_init(PACKAGE_BUGREPORT, CPL_MSG_WARNING);


    /* Verify the test data */
    cpl_test_eq(sizeof(off_x_init), NFRAMES * sizeof(off_x_init[0]));
    cpl_test_eq(sizeof(off_y_init), sizeof(off_x_init));
    cpl_test_eq(sizeof(psigmas),    NSIGMAS * sizeof(psigmas[0]));

    if (cpl_msg_get_level() <= CPL_MSG_INFO) {
        cpl_geom_img_offset_saa_bench(CPL_GEOM_FIRST, 10, 16, 4*IMAGESZ,
                                      4*IMAGESZ, 0);
        cpl_geom_img_offset_saa_bench(CPL_GEOM_FIRST,  6, 18, 4*IMAGESZ,
                                      4*IMAGESZ, 1);
    }


    /* Create an image set */
    img = cpl_image_fill_test_create(IMAGESZ, IMAGESZ);
    iset = cpl_imagelist_new();
    cpl_imagelist_set(iset, img, 0);

    for (i=1; i < NFRAMES; i++) {
        cpl_image * copy = cpl_image_duplicate(img);

        cpl_image_shift(copy, (int)(-off_x_init[i]), (int)(-off_y_init[i]));
        cpl_imagelist_set(iset, copy, i);
    }

    cpl_test_eq(cpl_imagelist_get_size(iset), NFRAMES);
    
    /* Not modified */
    off_vec_x = cpl_vector_wrap(NFRAMES, (double*)off_x_init);
    off_vec_y = cpl_vector_wrap(NFRAMES, (double*)off_y_init);
    offs_est = cpl_bivector_wrap_vectors(off_vec_x, off_vec_y);

    sigmas = cpl_vector_wrap(NSIGMAS, (double*)psigmas); /* Not modified */

    cpl_test_error(CPL_ERROR_NONE);

    /* Get some cross-correlation apertures */
    cpl_msg_info("","Apply the apertures detection");
    aperts = cpl_apertures_extract(cpl_imagelist_get(iset, 0), sigmas, &pos);
    cpl_vector_unwrap(sigmas);

    cpl_test_nonnull(aperts);

    naperts = cpl_apertures_get_size(aperts);

    cpl_test_leq(1, naperts);

    aperts_pos = cpl_bivector_new(naperts);
    aperts_pos_x = cpl_bivector_get_x(aperts_pos);
    aperts_pos_y = cpl_bivector_get_y(aperts_pos);
    for (i=0; i<naperts; i++) {
        cpl_vector_set(aperts_pos_x, i, cpl_apertures_get_max_x(aperts, i+1));
        cpl_vector_set(aperts_pos_y, i, cpl_apertures_get_max_y(aperts, i+1));
    }
    cpl_apertures_delete(aperts);

    cpl_test_error(CPL_ERROR_NONE);
   
    /* Refine the offsets with cpl_geom_img_offset_fine */
    cpl_msg_info("","Refine the offsets");
    correl = cpl_vector_new(NFRAMES);

    offs_ref = cpl_geom_img_offset_fine(iset, offs_est, aperts_pos,
                                        15, 15, 15, 15, correl);

    cpl_test_nonnull(offs_ref);

    cpl_test_eq(cpl_bivector_get_size(offs_ref), NFRAMES);

    cpl_vector_delete(correl);
    cpl_bivector_unwrap_vectors(offs_est);
    cpl_vector_unwrap(off_vec_x);
    cpl_vector_unwrap(off_vec_y);
    cpl_bivector_delete(aperts_pos);

    x_offs = cpl_bivector_get_x_data(offs_ref);
    y_offs = cpl_bivector_get_y_data(offs_ref);
    for (i=0; i < cpl_bivector_get_size(offs_ref); i++) {
        cpl_test_abs(x_offs[i], off_x_init[i], MAX_SHIFT_ERROR2);
        cpl_test_abs(y_offs[i], off_y_init[i], MAX_SHIFT_ERROR2);
    }

    for (i = 0; i < (int)(sizeof(geoms)/sizeof(geoms[0])); i++) {
        const cpl_geom_combine geom = geoms[i];
        const int rejmin = 1;
        const int rejmax = 1;
        const int maximg = NFRAMES - rejmin - rejmax;

        /* Shift and add */
        cpl_msg_info("", "Shift and add with geom number %d", (int)geom);
        combined = cpl_geom_img_offset_saa(iset, offs_ref,  CPL_KERNEL_DEFAULT,
                                           rejmin, rejmax, geom, NULL, NULL);

        cpl_test_nonnull(combined);

        if (combined == NULL) continue;

        cpl_test_eq(cpl_image_get_type(combined[1]), CPL_TYPE_INT);

        if (cpl_image_get_min(combined[1]) == 0) {
            cpl_test(cpl_image_count_rejected(combined[0]));
        } else {
            cpl_test_leq(1, cpl_image_get_min(combined[1]));
            cpl_test_zero(cpl_image_count_rejected(combined[0]));
        }

        cpl_test_eq(cpl_image_get_size_x(combined[0]),
                    cpl_image_get_size_x(combined[1]));

        cpl_test_eq(cpl_image_get_size_y(combined[0]),
                    cpl_image_get_size_y(combined[1]));

        cpl_test_leq(cpl_image_get_max(combined[1]), maximg);

        if (geom == CPL_GEOM_INTERSECT) {
            cpl_test_eq(cpl_image_get_max(combined[1]), maximg);
            cpl_test_leq(1, cpl_image_get_min(combined[1]));
        } else if (geom == CPL_GEOM_FIRST) {
            /* FIXME: Should at least be 1 */
            cpl_test_leq(0, cpl_image_get_min(combined[1]));
        } else if (geom == CPL_GEOM_UNION) {
            cpl_test_leq(0, cpl_image_get_min(combined[1]));
        }

        cpl_msg_info("", "Minimum value in contribution map: %g",
                     cpl_image_get_min(combined[1]));

        cpl_image_delete(combined[0]);
        cpl_image_delete(combined[1]);
        cpl_free(combined);

    }


    cpl_bivector_delete(offs_ref);
    img = cpl_imagelist_unset(iset, 0);
    cpl_imagelist_delete(iset);


    /* Shift and add of two uniform images - with no offsets */

    iset = cpl_imagelist_new();
    cpl_imagelist_set(iset, img, 0);
    cpl_image_threshold(img, 1.0, 1.0, 1.0, 1.0);
    cpl_image_accept_all(img);
    img = cpl_image_duplicate(img);
    cpl_imagelist_set(iset, img, 1);

    off_vec_x = cpl_vector_new(2);
    cpl_vector_fill(off_vec_x, 0.0);
    offs_ref = cpl_bivector_wrap_vectors(off_vec_x,
                                         cpl_vector_duplicate(off_vec_x));

    if (cpl_msg_get_level() <= CPL_MSG_DEBUG)
        cpl_plot_image("","","", img);

    for (i = 0; i < (int)(sizeof(geoms)/sizeof(geoms[0])); i++) {
        const cpl_geom_combine geom = geoms[i];
        double pos_x, pos_y;

        cpl_msg_info("", "Shift and add with geom number %d", (int)geom);

        combined = cpl_geom_img_offset_saa(iset, offs_ref,  CPL_KERNEL_DEFAULT,
                                           0, 0, geom, &pos_x, &pos_y);

        cpl_test_nonnull(combined);

        if (combined == NULL) continue;

        cpl_test_eq(cpl_image_get_size_x(combined[0]), IMAGESZ);
        cpl_test_eq(cpl_image_get_size_x(combined[0]), IMAGESZ);

        cpl_test_eq(cpl_image_get_type(combined[1]), CPL_TYPE_INT);

        cpl_test_eq(cpl_image_get_size_x(combined[0]),
                    cpl_image_get_size_x(combined[1]));

        cpl_test_eq(cpl_image_get_size_y(combined[0]),
                    cpl_image_get_size_y(combined[1]));

        if (cpl_image_get_min(combined[1]) == 0) {
            cpl_test(cpl_image_count_rejected(combined[0]));
        } else {
            cpl_test_leq(1, cpl_image_get_min(combined[1]));
            cpl_test_zero(cpl_image_count_rejected(combined[0]));
        }

        cpl_test_eq(cpl_image_get_max(combined[1]), 2);

        if (geom == CPL_GEOM_INTERSECT) {
            cpl_test_eq(cpl_image_get_max(combined[1]), 2);
            /* FIXME: Should at least be 1 */
            cpl_test_leq(0, cpl_image_get_min(combined[1]));
        } else if (geom == CPL_GEOM_FIRST) {
            /* FIXME: Minimum value is zero */
            cpl_test_leq(0, cpl_image_get_min(combined[1]));
        } else if (geom == CPL_GEOM_UNION) {
            cpl_test_leq(0, cpl_image_get_min(combined[1]));
        }

#ifdef TEST_RESAMPLING
        /* Resampling introduces noise at the edge */
        /* NB: Comparison works for all modes, due to zero offset ... */
        cpl_test_image_abs(combined[0], img, MAX_SHIFT_ERROR2);
#endif

        if (cpl_msg_get_level() <= CPL_MSG_DEBUG) {
            cpl_image_subtract(combined[0], img);
            cpl_plot_image("","","", combined[0]);
        }

        cpl_image_delete(combined[0]);
        cpl_image_delete(combined[1]);
        cpl_free(combined);


        /* Now try to combine two images, the second shifted along the X-axis */

        cpl_image_shift(img, -MAX_SHIFT_ERROR1, 0);
        cpl_image_accept_all(img);

        cpl_vector_set(off_vec_x, 1, MAX_SHIFT_ERROR1);

        combined = cpl_geom_img_offset_saa(iset, offs_ref, CPL_KERNEL_DEFAULT,
                                           0, 0, geom, &pos_x, &pos_y);

        cpl_test_nonnull(combined);

        if (combined == NULL) continue;

        cpl_test_eq(cpl_image_get_max(combined[1]), 2);

        if (cpl_image_get_min(combined[1]) == 0) {
            cpl_test(cpl_image_count_rejected(combined[0]));
        } else {
            cpl_test_zero(cpl_image_count_rejected(combined[0]));
        }

        cpl_test_eq(cpl_image_get_size_y(combined[0]), IMAGESZ);

        if (geom == CPL_GEOM_INTERSECT) {

#ifdef SAVE_COMBINED
            cpl_image_save(combined[0], "PI.fits", CPL_BPP_IEEE_DOUBLE,
                           NULL, CPL_IO_CREATE);
            cpl_image_save(combined[1], "CI.fits", CPL_BPP_8_UNSIGNED,
                           NULL, CPL_IO_CREATE);
#endif

            cpl_test_eq(cpl_image_get_size_x(combined[0]), IMAGESZ
                        - MAX_SHIFT_ERROR1);

        } else if (geom == CPL_GEOM_FIRST) {

#ifdef SAVE_COMBINED
            cpl_image_save(combined[0], "PF.fits", CPL_BPP_IEEE_DOUBLE,
                           NULL, CPL_IO_CREATE);
            cpl_image_save(combined[1], "CF.fits", CPL_BPP_8_UNSIGNED,
                           NULL, CPL_IO_CREATE);
#endif


            cpl_test_eq(cpl_image_get_size_x(combined[0]), IMAGESZ);

        } else if (geom == CPL_GEOM_UNION) {
            cpl_test_eq(cpl_image_get_size_x(combined[0]), IMAGESZ
                        + MAX_SHIFT_ERROR1);

#ifdef SAVE_COMBINED
            cpl_image_save(combined[0], "PU.fits", CPL_BPP_IEEE_DOUBLE,
                           NULL, CPL_IO_CREATE);
            cpl_image_save(combined[1], "CU.fits", CPL_BPP_8_UNSIGNED,
                           NULL, CPL_IO_CREATE);
#endif

            cpl_test_eq(cpl_image_get_min(combined[1]), 1);
        }


        img = cpl_imagelist_get(iset, 0);

        if (cpl_msg_get_level() <= CPL_MSG_DEBUG) {
            cpl_plot_image("","","", combined[0]);

            if (geom == CPL_GEOM_FIRST) {
                cpl_image_subtract(combined[0], img);
                cpl_plot_image("","","", combined[0]);
            }
        }

        cpl_image_delete(combined[0]);
        cpl_image_delete(combined[1]);
        cpl_free(combined);

        /* Reset offset and 2nd image */
        cpl_vector_fill(off_vec_x, 0.0);
        img = cpl_image_duplicate(img);
        cpl_imagelist_set(iset, img, 1);

    }

    cpl_imagelist_delete(iset);
    cpl_bivector_delete(offs_ref);

    return cpl_test_end(0);
}

/**@}*/

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief      Benchmark the CPL function
  @param mode CPL_GEOM_INTERSECT, CPL_GEOM_UNION, CPL_GEOM_FIRST
  @param nr   The number of repeats
  @param nz   The number of planes
  @param nx   The image X-size
  @param ny   The image Y-size
  @param no   The number of outlier pixels to ignore (both min and max)
  @return    void
 */
/*----------------------------------------------------------------------------*/
void cpl_geom_img_offset_saa_bench(cpl_geom_combine mode, int nr, int nz,
                                   int nx, int ny, int no)
{

    cpl_bivector  * offset  = cpl_bivector_new(nz);
    cpl_vector    * off_x   = cpl_bivector_get_x(offset);
    cpl_vector    * off_y   = cpl_bivector_get_y(offset);
    cpl_image     * imgd    = cpl_image_fill_test_create(nx, ny);
    cpl_image     * img     = cpl_image_cast(imgd, CPL_TYPE_FLOAT);
    cpl_imagelist * imglist = cpl_imagelist_new();
    cpl_image    ** combined;
    double          cputime = 0.0;
    cpl_flops       nflops = 0;

    int ir, iz;

    cpl_test_leq(1, nz);

    /* Create bivector of shifts, from 0.4 to 0.6 of the pixel range */
    /* Create list of shifted images */
    cpl_vector_set(off_x, 0, 0.0);
    cpl_vector_set(off_y, 0, 0.0);

    cpl_image_delete(imgd);

    cpl_imagelist_set(imglist, img, 0);


    for (iz = 1; iz < nz; iz++) {
        cpl_image * copy = cpl_image_duplicate(img);
        const int dx = (int)(0.1 * nx - 0.2 * nx * cpl_drand());
        const int dy = (int)(0.1 * ny - 0.2 * ny * cpl_drand());

        cpl_vector_set(off_x, iz, (double)dx);
        cpl_vector_set(off_y, iz, (double)dy);

        cpl_image_shift(copy, -dx, -dy);
        cpl_image_accept_all(copy);
        cpl_imagelist_set(imglist, copy, iz);
    }

    if (cpl_msg_get_level() <= CPL_MSG_DEBUG)
        cpl_bivector_dump(offset, stdout);

    for (ir = 0; ir < nr; ir++) {
        const cpl_flops flops  = cpl_tools_get_flops();
        const double secs = cpl_test_get_cputime();

        combined = cpl_geom_img_offset_saa(imglist, offset, CPL_KERNEL_DEFAULT,
                                           no, no, mode, NULL, NULL);

        cputime += cpl_test_get_cputime() - secs;
        nflops += cpl_tools_get_flops() - flops;

        if (combined == NULL) continue;

        cpl_image_delete(combined[0]);
        cpl_image_delete(combined[1]);
        cpl_free(combined);
    }

    cpl_msg_info(cpl_func, "Time to benchmark with mode=%d, nr=%d, nz=%d, "
                 "nx=%d, ny=%d, no=%d [s]: %g (%g MFLOP/s)",
                 mode, nr, nz, nx, ny, no, cputime,
                 cputime > 0.0 ? (double)nflops/cputime/1e6 : 0.0);

    cpl_bivector_delete(offset);
    cpl_imagelist_delete(imglist);

    return;
}
